from typing import Dict, List, Literal, Optional, Union
from pydantic import validator, constr, BaseModel, validate_arguments
from docarray import BaseDoc, DocList
from wasabi import msg
from pathlib import Path
import json
import re
from .config import registry




class Token(BaseDoc):
    id: Optional[Union[int, str]] = None
    text: constr(strip_whitespace=True, min_length=1)
    space: Optional[bool] = False
    indices: Optional[List[int]] = None
    is_pad: Optional[bool] = False
    
    @validator('id')
    def validate_id(cls, v, values, **kwargs):
        return str(v)
    

class TokenList(DocList[Token]):
    
    raw_text: Optional[str] = None
    processed_text: Optional[str] = None
    
    def char_to_token(self, char_index: int):
        """字符索引转换为token索引
        """
        for idx, doc in enumerate(self):
            if doc.indices and char_index in doc.indices:
                return idx
        return None
    
    def token_to_char(self, token_index: int) -> Optional[List[int]]:
        """token索引转换为字符索引
        """
        if token_index >= len(self):
            return None
        return self[token_index].indices
    
    @property
    def pad_length(self):
        return sum([1 for token in self if token.is_pad])
    
    @property
    def pad_mask(self):
        mask = []
        for token in self:
            if token.is_pad:
                mask.append(0)
            else:
                mask.append(1)
        return mask
    
    @property
    def ids(self):
        return [int(token.id) for token in self]
    
    def check_indices_in_processed(self):
        for token in self:
            if token.indices:
                assert token.text == self.processed_text[token.indices[0]: token.indices[-1]+1], f"token {token.text} indices error"
        msg.good("token indices check pass")
        
    
    def check_indices_in_raw(self):
        for token in self:
            if token.indices:
                assert token.text == self.raw_text[token.indices[0]: token.indices[-1]+1], f"token {token.text} indices error"
        msg.good("token indices check pass")
    


class Tokenizer(BaseModel):
    
    max_length: int
    vocab: Union[str, Dict[str, int], Path]
    pad_side: Literal['left', 'right'] = "left"
    unk_token: str = "UNK"
    pad_id: int = 0
    
    
    @validator('vocab')
    def validate_vocab(cls, v, values, **kwargs):
        if isinstance(v, str) or isinstance(v, Path):
            with open(v, 'r', encoding='utf-8') as f:
                v = json.load(f)
        if 'unk_token' in values:
            assert values['unk_token'] in v, f"unk_token {values['unk_token']} not in vocab"
        return v
    
    @validator('unk_token')
    def validate_unk_token(cls, v, values, **kwargs):
        if 'vocab' in values:
            assert v in values['vocab'], f"unk_token {v} not in vocab"
        return v
    
    @property
    def token2id(self):
        return self.vocab
    
    @property
    def id2token(self):
        return {v: k for k, v in self.vocab.items()}
    
    @property
    def pad_token(self):
        return self.id2token[self.pad_id]
    
    @property
    def unk_token_id(self):
        return self.token2id[self.unk_token]
    
    def encode(self, tokens: TokenList) -> TokenList:
        for token in tokens:
            if token.text in self.vocab:
                token.id = self.vocab[token.text]
            else:
                token.id = self.unk_token_id
        return tokens
    
    def encode_batch(self, batch_tokens: List[TokenList]) -> List[TokenList]:
        return [self.encode(tokens) for tokens in batch_tokens]
    
    
    def decode(self, ids: List[int]) -> List[str]:
        return [self.id2token[int(id)] for id in ids]
    
    
    def pad(self, tokens: TokenList):
        pad_id = self.pad_id
        if self.pad_side == "left":
            pad_tokens = TokenList([Token(text=self.pad_token, id=pad_id, is_pad=True)] * (self.max_length - len(tokens)))
            pad_tokens.extend(tokens)
            pad_tokens.raw_text = tokens.raw_text
            pad_tokens.processed_text = tokens.processed_text
            tokens = pad_tokens
        else:
            tokens.extend([Token(text=self.pad_token, id=pad_id, is_pad=True)] * (self.max_length - len(tokens)))
        return tokens
    
        
    def pad_batch(self, batch_tokens: List[TokenList]):
        batch_tokens = [self.pad(tokens) for tokens in batch_tokens]
        return batch_tokens
        
    
    def save_vocab_json(self, save_path: str):
        """保存词表json文件

        Args:
            save_path (str): 保存路径
        """
        with open(save_path, 'w', encoding='utf-8') as f:
            json.dump(self.vocab, f)
            
            
    def save_vocab_txt(self, save_path: str):
        """保存词表txt文件

        Args:
            save_path (str): 保存路径
        """
        with open(save_path, 'w', encoding='utf-8') as f:
            for k, v in self.vocab.items():
                f.write(f'{k} {v}\n')
                
    def to_disk(self, save_path: Union[str, Path]):
        """保存词表json文件

        Args:
            save_path (Union[str, Path]): 保存路径
        """
        json_str = self.json()
        with open(save_path, 'w', encoding='utf-8') as f:
            f.write(json_str)
    
    
    @classmethod     
    def from_disk(cls, load_path: Union[str, Path]):
        """从磁盘加载词表

        Args:
            load_path (Union[str, Path]): 加载路径
        """
        with open(load_path, 'r', encoding='utf-8') as f:
            json_str = f.read()
        return cls(**json.loads(json_str))
    
            
    
    def preprocess(self, text: str) -> TokenList:
        raise NotImplementedError
        
    
    def tokenize(self, tokens: TokenList) -> TokenList:
        raise NotImplementedError
    
   
    @validate_arguments
    def __call__(self, batch_text: Union[List[str], str], padding: bool = True) -> List[TokenList]:
        """对文本进行预处理,分词,编码,填充

        Args:
            batch_text (Union[List[str], str]): 输入文本
            padding (bool, optional): 是否填充. Defaults to True.

        Returns:
            List[TokenList]: 分词后的结果
        """
        if isinstance(batch_text, str):
            batch_text = [batch_text]
        batch_tokens: List[TokenList] = [self.preprocess(text) for text in batch_text]
        batch_tokens = [self.tokenize(tokens) for tokens in batch_tokens]
        batch_tokens = self.encode_batch(batch_tokens)
        if padding:
            batch_tokens = self.pad_batch(batch_tokens)
        return batch_tokens
        

class CharTokenizer(Tokenizer):
    """非常简单的基于字符的分词器,用于中文
    """
    remove_space: bool = True
    
    def preprocess(self, text: str) -> TokenList:
        tokens = TokenList()
        tokens.raw_text = text
        processed_text = text.strip()
        if self.remove_space:
            processed_text = processed_text.replace(' ', '')
        tokens.processed_text = processed_text
        return tokens
        
    def tokenize(self, tokens: TokenList) -> TokenList:
        """对文本进行字符切分

        Args:
            text (str): 输入文本

        """
        for i, char in enumerate(tokens.processed_text):
            if len(tokens) == self.max_length:
                break
            if not char.isspace():
                tokens.append(Token(text=char, indices=[i], space=False))
            else:
                tokens[-1].space = True
        return tokens


@registry.tokenizers.register('char')
def create_char_tokenizer(vocab: Union[Dict[str, int], str, Path], 
                          max_length: int, 
                          pad_side: Literal['left', 'right'] = "left", 
                          unk_token: str = "UNK", 
                          pad_id: int = 0, 
                          remove_space: bool = True):
    
    return CharTokenizer(vocab=vocab, 
                         max_length=max_length, 
                         pad_side=pad_side, 
                         unk_token=unk_token, 
                         pad_id=pad_id, 
                         remove_space=remove_space)
    

class WordTokenizer(Tokenizer):
    """基于空格的分词器,主要用于英文
    """
    
    do_lower: bool = True # 是否小写 My -> my
    split_number: bool = True # 是否切分数字 911 -> 9 1 1
    split_apostrophe: bool = True  # 是否切分撇号 don't -> don 't
    split_punctuation: bool = True # 是否切分标点符号 i am. -> i am .
    
    def preprocess(self, text: str) -> TokenList:
        tokens = TokenList()
        tokens.raw_text = text
        text = text.strip()
        # 将多个空格替换为一个空格
        text = re.sub(r'\s+', ' ', text)
        if self.do_lower:
            text = text.lower()
        tokens.processed_text = text
        return tokens
    
    def tokenize(self, tokens: TokenList) -> TokenList:
        spans = tokens.processed_text.split(' ')
        for i, span in enumerate(spans):
            if i == len(spans) - 1:
                space = False
            else:
                space = True
            if self.split_number:
                # 把数字单独拆分出来 911 -> 9 1 1
                # 切分数字
                _token = re.sub(r'(\d)', r' \1 ', span).split(' ')
                _token = [s for s in _token if s]
                # 如果切分后的长度大于1,说明有数字
                for s in _token[:-1]:
                    tokens.append(Token(text=s, space=False))
                    
                if _token[-1].isdigit():
                    tokens.append(Token(text=_token[-1], space=space))
                    
                else:
                    if self.split_apostrophe:
                        # 切分撇号 don't -> do n't
                        _token1 = re.sub(r"([a-zA-Z])'([a-zA-Z])", r"\1 '\2", _token[-1]).split(' ')
                        for s in _token1[:-1]:
                            if s:
                                tokens.append(Token(text=s, space=False))
                        tokens.append(Token(text=_token1[-1], space=space))
            else:
                if self.split_apostrophe:
                    _token = re.sub(r"([a-zA-Z])'([a-zA-Z])", r"\1 '\2", span).split(' ')
                    for s in _token[:-1]:
                        if s:
                            tokens.append(Token(text=s, space=False))
                    tokens.append(Token(text=_token[-1], space=space))
                else:
                    tokens.append(Token(text=span, space=space))
        if self.split_punctuation:
            # 切分标点符号
            new_tokens = TokenList()
            for token in tokens:
                _token = re.sub(r"([a-zA-Z])([,.!?])", r"\1 \2", token.text).split(' ')
                _token = [s for s in _token if s]
                for s in _token[:-1]:
                    new_tokens.append(Token(text=s, space=False))
                new_tokens.append(Token(text=_token[-1], space=token.space))
            new_tokens.raw_text = tokens.raw_text
            new_tokens.processed_text = tokens.processed_text
            tokens = new_tokens
        for i, token in enumerate(tokens):
            if i == 0:
                token.indices = list(range(len(token.text)))
            else:
                if tokens[i-1].space:
                    start = tokens[i-1].indices[-1] + 2
                    end = start + len(token.text)
                    token.indices = list(range(start, end))
                else:
                    start = tokens[i-1].indices[-1] + 1
                    end = start + len(token.text)
                    token.indices = list(range(start, end))
        final_tokens = tokens[:self.max_length]
        final_tokens.raw_text = tokens.raw_text
        final_tokens.processed_text = tokens.processed_text
        return final_tokens
    

@registry.tokenizers.register('word')
def create_word_tokenizer(vocab: Union[Dict[str, int], str, Path], 
                          max_length: int, 
                          pad_side: Literal['left', 'right'] = "left", 
                          unk_token: str = "UNK",
                          pad_id: int = 0,
                          do_lower: bool = True,
                          split_number: bool = True,
                          split_apostrophe: bool = True,
                          split_punctuation: bool = True):
    
    return WordTokenizer(vocab=vocab, 
                         max_length=max_length, 
                         pad_side=pad_side, 
                         unk_token=unk_token, 
                         pad_id=pad_id, 
                         do_lower=do_lower, 
                         split_number=split_number, 
                         split_apostrophe=split_apostrophe, 
                         split_punctuation=split_punctuation)
                          


from confection import Config
import confection
import catalogue


class registry(confection.registry):
    
    tokenizers = catalogue.create("nlu", "tokenizers", entry_points=True)
    inferences = catalogue.create("nlu", "inferences", entry_points=True)
    languages = catalogue.create("nlu", "languages", entry_points=True)
    
    

__all__ = ["registry", "Config"]


import psutil
from typing import Callable, List, Tuple, Optional, Union
from wasabi import msg
import time
from .tokenizer import TokenList
from pathlib import Path
from typing import Literal, Optional
from .config import Config
from contextlib import contextmanager


def get_cpu_memory(tag: str):
    """获取函数运行时的cpu内存

    Args:
        tag (str): 标签
    """
    def out_wrapper(fn: Callable):
        def wrapper(*args, **kwargs):
            p = psutil.Process()
            cpu_memory = p.memory_info().rss
            outputs = fn(*args, **kwargs)
            cpu_memory = p.memory_info().rss - cpu_memory
            msg.good(f"{tag} cpu memory: {cpu_memory / 1024 / 1024}MB")
            return outputs
        return wrapper
    return out_wrapper

def get_spent_time(tag: str):
    """获取函数运行时的时间

    Args:
        tag (str): 标签
    """
    def out_wrapper(fn: Callable):
        def wrapper(*args, **kwargs):
            start_time = time.time()
            outputs = fn(*args, **kwargs)
            end_time = time.time()
            spent_time = round(end_time - start_time, 4)
            msg.info(f"{tag} spent time: {spent_time}s")
            return outputs
        return wrapper
    return out_wrapper

@contextmanager
def print_spent_time(tag: str):
    start_time = time.time()
    yield
    end_time = time.time()
    spent_time = round(end_time - start_time, 4)
    msg.info(f"{tag} spent time: {spent_time}s")



def get_ents(tags: List[str]) -> List[Tuple[int, int, str]]:
    """从序列标签中提取实体

    Args:
        tags (List[str]): 序列标签.

    Returns:
        List[Tuple[int, int, str]]: 实体列表.例如, [(2, 6, 'PER')]
    """
    entities = []
    entity = []
    for i, tag in enumerate(tags):
        if tag.startswith('B-'):
            if entity:
                entities.append(tuple(entity))
            entity = [i, i + 1, tag.split('-')[1]]
        elif tag.startswith('I-'):
            if entity:
                if entity[2] == tag.split('-')[1]:
                    entity[1] = i + 1
        else:
            if entity:
                entities.append(tuple(entity))
            entity = []
    if len(entity) == 3:
        entities.append(tuple(entity))
    return entities
        
    
def ner_post_process(labels: List[str], tokens: Optional[TokenList] = None) -> List[Tuple[int, int, str, str]]:
    """对ner模型预测的token标签进行后处理，得到最终的实体

    Args:
        labels (List[str]): 模型预测的序列标签.
        text (str): 原始文本.

    Returns:
        List[Tuple[int, int, str, str]]: 实体列表.例如, [(2, 6, '明天上午', 'DateChanged')]
    """
    if tokens:
        assert len(labels) == len(tokens)
    ents = get_ents(labels)
    if not tokens:
        return ents
    results = []
    for ent in ents:
        start, end, label = ent
        start_char = tokens.token_to_char(start)
        if start_char:
            start_char = start_char[0]
        end_char = tokens.token_to_char(end - 1)
        if end_char:
            end_char = end_char[-1] + 1
        if start_char is not None and end_char is not None:
            results.append((start_char, end_char, tokens.processed_text[start_char:end_char], label))
    return results


def get_nlu_config(domain: str, 
                   lang: Literal["cmn", "zho", "eng"], 
                   tokenzer_type: Literal["char", "word"],
                   max_length: int,
                   pad_side: Literal["left", "right"],
                   pad_id: int = 0,
                   unk_token: str = "UNK",
                   ) -> str:
    
    config_str = """
    [nlu]
    @languages = cmn
    domain = "schedule"

    [nlu.domain_inference]
    @inferences = "domain.fasttext"
    model_path = ${path.domain}

    [nlu.intention_inference]
    @inferences = "intention.fasttext"
    model_path = ${path.intention}

    [nlu.ner_inference]
    @inferences = "ner.onnx.cloud"
    model_path = ${path.ner}
    label_path = ${path.label}

    [nlu.tokenizer]
    @tokenizers = "char"
    max_length = 32
    vocab = ${path.vocab}
    pad_side = "left"
    pad_id = 0
    unk_token = "UNK"

    [path]
    ner = "model/ner/model.onnx"
    vocab = "model/ner/word2id.json"
    label = "model/ner/id2label.json"
    domain = "model/domain/model.bin"
    intention = "model/intention/model.bin"
    """
    overrides = {"nlu.domain": domain, "nlu.@languages": lang, "nlu.tokenizer.@tokenizers": tokenzer_type, "nlu.tokenizer.max_length": max_length, "nlu.tokenizer.pad_side": pad_side, "nlu.tokenizer.pad_id": pad_id, "nlu.tokenizer.unk_token": unk_token}
    config = Config().from_str(config_str, overrides=overrides, interpolate=False)
    return config
    


from pydantic import BaseModel
from typing import Literal, List, Dict, Optional

    
class NLUInput(BaseModel):
    rawText: str
    
    
NERResult = Dict[str, str]
    
    
    
class CLSResult(BaseModel):
    label: str 
    score: float
         
         
class ParserResult(BaseModel):
    domain: str
    intention: str
    ner:  Dict[str, str]
    

class Slot(BaseModel):
    raw_value: str
    value: str
    
class Event(BaseModel):
    eventId: str
    raw: ParserResult
    score: float = 1.0
    slots: Dict[str, Slot]
    
    
class ParserOutput(BaseModel):
    code: Literal[0, 1] = 0
    data: List[Event]
    message: str = "success"
    
    
class HealthCheckOutput(BaseModel):
    code: Literal[0, 1] = 0
    message: str = "success"
    
    
class DomainResult(BaseModel):
    domain: str 
    probability: float
    session: Dict[str, str] = {}
    
    
class DomainOutput(BaseModel):
    code: Literal[0, 1] = 0
    data: DomainResult
    message: str = "success"


from jsonargparse import CLI 
from .io import ParserOutput, HealthCheckOutput, DomainOutput, NLUInput
from .language import NLU
import uvicorn
from .utils import print_spent_time
from wasabi import msg



def serve(checkpoint: str, example: str = '你好', port: int = 18000, host: str = "localhost"):
    
    nlu = NLU.from_checkpoint(checkpoint)
    
    domain = nlu.domain
    
    
    msg.info(f"loading model {domain}")
    
    from fastapi import FastAPI
    app = FastAPI(description=f"{domain} Inference API")
    
    @app.post("/parser", response_model=ParserOutput)
    def parse(inputs: NLUInput):
        
        with print_spent_time("parser"):
            output: ParserOutput = nlu.predict_parser(inputs.rawText)
            
        return output
    
    
    @app.post("/domain", response_model=DomainOutput)
    def domain(inputs: NLUInput):
        
        with print_spent_time("domain"):
            output =  nlu.predict_domain(inputs.rawText)
        return output
    
    
    @app.post("/health", response_model=HealthCheckOutput)
    def health():
        try:
            domain_results = nlu.predict_domain(example)
            parser_results = nlu.predict_parser(example)
            return HealthCheckOutput(code=0, message="success")
        except Exception as e:
            return HealthCheckOutput(code=1, message=str(e))
        
    @app.get("/health")
    def health():
        try:
            domain_results = nlu.predict_domain(example)
            parser_results = nlu.predict_parser(example)
            return {"code": 0, "message": "success"}
        except Exception as e:
            return {"code": 1, "message": str(e)}
        
    uvicorn.run(app, host=host, port=port)


def run():
    CLI(serve)


from pydantic import BaseModel
from .tokenizer import Tokenizer, CharTokenizer, WordTokenizer
from .inference import CLSInference, NERInference
from .io import ParserOutput, DomainOutput, CLSResult, NERResult, DomainResult, Event, ParserResult, Slot
from .config import registry
from .config import Config
from typing import Optional
from pathlib import Path
from wasabi import msg


class NLU(BaseModel):
    
    domain: str
    tokenizer: Tokenizer
    domain_inference: Optional[CLSInference] = None
    intention_inference: Optional[CLSInference] = None
    ner_inference: Optional[NERInference] = None
    config: Optional[Config] = None
    version: Optional[str] = None
    
    class Config:
        arbitrary_types_allowed = True
    
    
    def predict_domain(self, text: str) -> DomainOutput:
        if self.domain_inference:
            tokens = self.tokenizer(text, padding=False)[0]
            result: CLSResult = self.domain_inference(tokens)
            data = DomainResult(domain=self.domain, probability=result.score)
            return DomainOutput(data=data, code=0, message="success")
        else:
            msg.warn(f"domain inference model not found for {self.domain}")
            result = CLSResult(label="unknown", score=0.0)
            data = DomainResult(domain=result.label, probability=result.score)
            return DomainOutput(data=data, code=1, message="domain inference model not found")
    
    def predict_parser(self, text: str) -> ParserOutput:
        tokens = self.tokenizer(text)[0]
        intention_result: CLSResult = self.intention_inference(tokens=tokens)
        if not self.ner_inference:
            msg.warn(f"ner inference model not found for {self.domain}")
            ner_result = {}
        else:
            ner_result: NERResult = self.ner_inference(tokens=tokens)
        parser_result = ParserResult(domain=self.domain, intention=intention_result.label, ner=ner_result)
        slots = {}
        for label in ner_result.keys():
            ent = ner_result[label]
            slots[label] = Slot(raw_value=ent, value=ent)
        event = Event(eventId=f"model_intention_{intention_result.label}", raw=parser_result, slots=slots)
        return ParserOutput(data=[event], code=0, message="success")    
    
    @classmethod
    def from_checkpoint(cls, checkpoint_dir: str):
        config_path = Path(checkpoint_dir) / "config.cfg"
        if not config_path.exists():
            raise FileNotFoundError(f"config file {config_path} not found")
        config = Config().from_disk(config_path)
        nlu = registry.resolve(config=config)['nlu']
        nlu.config = config
        return nlu
    
    def save_config(self, save_path: str):
        config_path = Path(save_path)
        self.config.to_disk(config_path, interpolate=False)
    

@registry.languages.register("cmn")
def create_cmn(domain: str, 
               tokenizer: CharTokenizer, 
               domain_inference: Optional[CLSInference] = None, 
               intention_inference: Optional[CLSInference] = None, 
               ner_inference: Optional[NERInference] = None,
               version: Optional[str] = None):
    return NLU(domain=domain, 
               tokenizer=tokenizer, 
               domain_inference=domain_inference, 
               intention_inference=intention_inference, 
               ner_inference=ner_inference,
               version=version)

@registry.languages.register("zho")
def create_zho(domain: str, 
               tokenizer: CharTokenizer, 
               domain_inference: Optional[CLSInference] = None, 
               intention_inference: Optional[CLSInference] = None, 
               ner_inference: Optional[NERInference] = None,
               version: Optional[str] = None):
    return NLU(domain=domain, 
               tokenizer=tokenizer, 
               domain_inference=domain_inference, 
               intention_inference=intention_inference, 
               ner_inference=ner_inference,
               version=version)

@registry.languages.register("eng")
def create_eng(domain: str, 
               tokenizer: WordTokenizer, 
               domain_inference: Optional[CLSInference] = None, 
               intention_inference: Optional[CLSInference] = None, 
               ner_inference: Optional[NERInference] = None,
               version: Optional[str] = None):
    return NLU(domain=domain, 
               tokenizer=tokenizer, 
               domain_inference=domain_inference, 
               intention_inference=intention_inference, 
               ner_inference=ner_inference,
               version=version)
        
    


from typing import Any, List, Tuple, Literal
from onnxruntime import InferenceSession, SessionOptions
from fasttext import load_model
from pathlib import Path
from .utils import ner_post_process
import json
import numpy as np
from .tokenizer import Tokenizer, TokenList
from pydantic import validate_arguments
from .config import registry
from .io import CLSResult, NERResult



class CLSInference():
    
    @validate_arguments
    def __call__(self, tokens: TokenList) -> CLSResult:
        tokens = self._preprocess(tokens)
        results = self._predict(tokens)
        results = self._postprocess(results)
        return results
    
    def _predict(self, tokens: TokenList) -> CLSResult:
        raise NotImplementedError
    
    def _preprocess(self, tokens: TokenList) -> TokenList:
        return tokens
    
    def _postprocess(self, result: CLSResult) -> CLSResult:
        return result



@registry.inferences.register("domain.fasttext")
class FasttextInference(CLSInference):
    def __init__(self, model_path: str):
        super().__init__()
        if isinstance(model_path, str):
            self.model = load_model(model_path)
        if isinstance(model_path, Path):
            self.model = load_model(str(model_path))
    
    def _predict(self, tokens: TokenList) -> CLSResult:
        """fasttext 模型推理

        Args:
            query (str): 原始问题文本
        """
        texts = [token.text for token in tokens if not token.is_pad]
        query = ' '.join(texts)
        result = self.model.predict(query)
        label = result[0][0]
        score = result[1][0]
        return CLSResult(label=label, score=round(score, 4))
    
    def _postprocess(self, result: CLSResult) -> CLSResult:
        result: CLSResult = CLSResult(label=result.label.replace("__label__", "").replace("#", ""), score=result.score)
        if result.label == "0":
            result.score = 0.0
        return result
    
@registry.inferences.register("intention.fasttext")
class FasttextInference(CLSInference):
    def __init__(self, model_path: str):
        super().__init__()
        if isinstance(model_path, str):
            self.model = load_model(model_path)
        if isinstance(model_path, Path):
            self.model = load_model(str(model_path))
    
    def _predict(self, tokens: TokenList) -> CLSResult:
        """fasttext 模型推理

        Args:
            query (str): 原始问题文本
        """
        texts = [token.text for token in tokens if not token.is_pad]
        query = ' '.join(texts)
        result = self.model.predict(query)
        label = result[0][0]
        score = result[1][0]
        return CLSResult(label=label, score=round(score, 4))
    
    def _postprocess(self, result: CLSResult) -> CLSResult:
        result = CLSResult(label=result.label.replace("__label__", "").replace("#", ""), score=result.score)
        return result

    
class NERInference():
    
    @validate_arguments
    def __call__(self, tokens: TokenList) -> NERResult:
        return self._predict(tokens)
    
    def _predict(self, tokens: TokenList) -> NERResult:
        raise NotImplementedError
    


@registry.inferences.register("ner.onnx.local")
class NERInferenceLocal(NERInference):
    """ner模型推理, 返回格式如下 (2, 6, '明天上午', 'DateChanged')
    """
    def __init__(self, 
                 embedding_path: str,
                 model_path: str,
                 label_path: str):
        super().__init__()
        so = SessionOptions()
        so.log_severity_level = 3
        if isinstance(model_path, str):
            self.classifier_model = InferenceSession(model_path, so)
        if isinstance(embedding_path, Path):
            self.classifier_model = InferenceSession(str(model_path), so)
        
        with open(label_path, 'r') as f:
            self.id2label = json.load(f)
            
        self.embedding_model = InferenceSession(embedding_path, so)
    
    
    def _predict(self, tokens: TokenList) -> NERResult:
        """ner 模型推理

        Args:
            query (str): 原始问题文本
        """
        embeddings = self.get_embeddings(ids=tokens.ids)
        labels = self.get_labels(embeddings)
        
        # 后处理
        ents = ner_post_process(labels, tokens=tokens)
        slots = {}
        for ent in ents:
            slots[ent[3]] = ent[2]
        return slots
    
    
    def get_embeddings(self, ids: List[int]) -> np.ndarray:
        """获取embedding

        Args:
            query (str): 原始问题文本
        """
        input_ids = np.array(ids, dtype=np.int64).reshape(1, -1)
        # 获取embedding
        embeddings = self.embedding_model.run(None, {'input': input_ids})[0]
        embeddings = np.array(embeddings, dtype=np.float32)
        return embeddings
    
    
    def get_labels(self, embeddings: np.ndarray) -> List[str]:
        """获取标签

        Args:
            embeddings (np.ndarray): embedding
        """
        label_ids = self.classifier_model.run(None, {'input': embeddings})[0][0]
        # 兼容tf版本转换的onnx模型
        if len(label_ids.shape) == 2:
            label_ids = np.argmax(label_ids, axis=-1)
            
        labels = [self.id2label[str(i)] for i in label_ids]
        return labels
    
    

@registry.inferences.register("ner.onnx.cloud")
class NERInferenceCloud(NERInference):
    """ner模型推理, 返回格式如下 (2, 6, '明天上午', 'DateChanged')
    """
    def __init__(self, 
                 model_path: str,
                 label_path: str):
        super().__init__()
        
        so = SessionOptions()
        so.log_severity_level = 3
        
        self.model = InferenceSession(str(model_path), so)
        
        with open(label_path, 'r') as f:
            self.id2label = json.load(f)
    
    def _predict(self, tokens: TokenList) -> NERResult:
        """ner 模型推理

        Args:
            query (str): 原始问题文本
        """
        label_ids = self.model.run(None, {'input': np.array([tokens.ids], dtype=np.int64)})[0][0]
        # 兼容tf版本转换的onnx模型
        if len(label_ids.shape) == 2:
            label_ids = np.argmax(label_ids, axis=-1)
            
        labels = [self.id2label[str(i)] for i in label_ids]
        
        # 后处理
        ents = ner_post_process(labels, tokens=tokens)
        
        slots = {}
        for ent in ents:
            slots[ent[3]] = ent[2]
        return slots
    
    
    